'
' User USB HID Routines
'
' Enumerates and keeps the USB attached using a low priority Timer1 interrupt.
'
    Dim wUSB_Timer1 As TMR1L.Word
    Dim PP0H As Byte System

    'Dim ReadTerminator          As Byte
   ' Dim DataAvailable           As Bit
'
' Set the Location of the Low Priority Hardware Interrupt (if in use)
'
$ifdef USB_SERVICE
    Dim bUSB_WREG_Save As Byte System
    Dim bUSB_BSR_Save As Byte System
    Dim bUSB_STATUS_Save As Byte System

    On_Low_Interrupt GoTo ISR_Service
$endif
(*
****************************************************************************
* Name    : mUSBService Macro                                              *
* Purpose : Service a USB connection. You need to use this routine         *
*         : periodically from your main program to maintain a connection   *
*         : (every 1 ms or so). You DO NOT need to call this routine if    *
*         : you have enabled the low priority interrupt handler (it is on  *
*         : by default)                                                    *
****************************************************************************
*)
$define mUSBService()   '
   USB_CheckBusStatus() '
   USB_DriverService()
(*
****************************************************************************
* Name    : EnableISR Macro                                                *
* Purpose : Enable the ISR to service USB connection                       *
****************************************************************************
*)
$define EnableISR()     '
   $ifdef USB_SERVICE   '
   INTCONbits_GIE = 1   '
   INTCONbits_GIEL = 1  '
   $endif
(*
****************************************************************************
* Name    : DisableISR Macro                                               *
* Purpose : Disable the ISR that services the USB connection               *
****************************************************************************
*)
$define DisableISR()    '
   $ifdef USB_SERVICE   '
   INTCONbits_GIE = 0   '
   INTCONbits_GIEL = 0  '
   $endif
(*
****************************************************************************
* Name    : DisableISR Macro                                               *
* Purpose : Disable the Low ISR that services the USB connection           *
****************************************************************************
*)
$define DisableLowISR() '
   $ifdef USB_SERVICE   '
   INTCONbits_GIEL = 0  '
   $endif
(*
****************************************************************************
* Name    : EnableLowISR Macro                                             *
* Purpose : Enable the Low ISR to service USB connection                   *
****************************************************************************
*)
$define EnableLowISR()  '
   $ifdef USB_SERVICE   '
   INTCONbits_GIEL = 1  '
   $endif
(*
****************************************************************************
* Name    : NoISRService Macro                                             *
* Purpose : service helper routine for maintain USB connection             *
*         : for some blocking calls                                        *
****************************************************************************
*)
$define NoISRService()  '
   $ifndef USB_SERVICE  '
   mUSBService()        '
   $endif
(*
****************************************************************************
* Name    : Initialise                                                     *
* Purpose : Initialise the USB module                                      *
****************************************************************************
*)
__Initialise:
    USB_InitialiseUSBDriver()
$ifdef USB_SERVICE
    wUSB_Timer1 = ReloadTimerValueEnum
    RCONbits_IPEN = 1       ' Enable priority levels on interrupts
    T1CON = 0               ' 1:1 Prescale value

    IPR1bits_TMR1IP = 0     ' Timer1 Overflow Interrupt Priority set Low
    PIR1bits_TMR1IF = 0     ' Clear the Interrupt Flag
    PIE1bits_TMR1IE = 1     ' Enable Timer1 interrupt
    T1CONbits_TMR1ON = 1    ' Switch Timer1 on
    INTCONbits_GIEL = 1     ' \
    INTCONbits_GIE = 1      ' / Enable global and peripheral interrupts
$endif
    GoTo OverHIDSubs        ' * Jump over the User HID subroutines *
(*
****************************************************************************
* Name    : ISR_Service                                                    *
* Purpose : Low Priority USB service event handler                         *
* Notes   : Uses a Timer1 overflow event as the interrupt source           *
****************************************************************************
*)
$ifdef USB_SERVICE
ISR_Service:
(*
    For some USB ports, enumeration must be performed very quickly in order to
    establish a connection - in some cases the standard USB_SERVICE_US option at
    around 1ms is too slow for enumeration to succeed.
    This routine will therefore use a very fast interrupt interval (USB_SERVICE_ENUM_US) to ensure enumeration is successful.
    Once attached, the slower USB_SERVICE_US value is used to maintain and serve the connection.
    If the device gets unplugged, the ISR reverts to USB_SERVICE_ENUM_US until the device is re-attached.

    Note that USB_SERVICE_ENUM_US will cause the ISR to fire constantly (which is what we want)
    but will leave very little MCU cycles for the main program until enumeration has completed.
*)
    bUSB_STATUS_Save = STATUS  ' \
    bUSB_WREG_Save = WREG      '   Save the state of STATUS, WREG, And BSR
    bUSB_BSR_Save = BSR        ' /
    Reset_Bank                 ' Reset the RAM bank so the code knows where it is
    $ifdef USB_CONNECT_LED
        Output USB_CONNECT_LED
    $endif
    If tUSB_Attached = True Then
        wUSB_Timer1 = wUSB_Timer1 + ReloadTimerValue
        $ifdef USB_CONNECT_LED
        Set USB_CONNECT_LED
        $endif
    Else
        wUSB_Timer1 = wUSB_Timer1 + ReloadTimerValueEnum
        $ifdef USB_CONNECT_LED
        Clear USB_CONNECT_LED
        $endif
    EndIf
    PIR1bits_TMR1IF = 0        ' Reset the Timer1 interrupt flag
    mUSBService()              ' Service the USB
    WREG = bUSB_WREG_Save      ' \
    BSR = bUSB_BSR_Save        '   Restore the state of WREG, BSR, and STATUS
    STATUS = bUSB_STATUS_Save  ' /
    Retfie                     ' Exit the interrupt
$endif

(*
****************************************************************************
* Name    : WaitForTX                                                      *
* Purpose : Blocks processing until USB TX buffer is available             *
****************************************************************************
*)
$define USB_WaitForTX() GoSub __WaitForTX

__WaitForTX:
    While tHID_TxIsBusy <> False
        If tUSB_Attached = False Then Return
        mUSBService()               ' Service the USB
    Wend
    Return
(*
****************************************************************************
* Name    : DataAvailable                                                  *
* Output  : WREG                                                           *
* Purpose : WREG is 1 if RX data is available, otherwise 0                 *
****************************************************************************
*)
$define HID_DataAvailable() _DataAvailable

_DataAvailable Macro- \Byte
    GoSub __DataAvailable
    Return_Wreg
Endm

#ifMacro- _DataAvailable
__DataAvailable:
    DisableLowISR()
    WREG = False
    If tHID_RxIsBusy = False Then
        WREG = True
    EndIf
    EnableLowISR()
    Return
#endIfMacro-
(*
*****************************************************************************
* Name    : HID_TxReport                                                    *
* Require : tHID_TxIsBusy() must return false. Value of bUSB_Length must    *
*         : be equal to or smaller than HID_INT_IN_EP_SIZE for an interrupt *
*         : endpoint, the largest buffer size is 64 bytes.                  *
* Input   : pBuffer = Pointer to where received bytes are to be stored      *
*         : pLength = Number of bytes to be transferred                     *
* Purpose : Use this macro to transfer data located in data memory.         *
*         : Note : tHID_TxIsBusy() must return false before user can call   *
*         : this function. Unexpected behavior will occur in this routine   *
*         : when tHID_TxIsBusy                                              *
*****************************************************************************
*)
$define HID_TxReport(pBuffer, pLength) '
    wUSB_FSR0 = VarPtr pBuffer         '
    WREG = pLength                     '
    _HIDTxReport

_HIDTxReport Macro-
    GoSub __HIDTxReport
Endm

#ifMacro- _HIDTxReport, _WriteReport, _WriteArray
__HIDTxReport:
    bUSB_Length = WREG
    If bUSB_Length > HID_INT_IN_EP_SIZE Then
        bUSB_Length = HID_INT_IN_EP_SIZE
    EndIf
    '
    ' Copy data from user's buffer to dual-RAM buffer
    '
    wUSB_FSR1 = VarPtr HID_Report_In
    WREG = bUSB_Length
    Inc WREG
HID_TxReportLoop:
    POSTINC1 = POSTINC0
    Decfsz WREG,f
    GoTo HID_TxReportLoop

    HID_BD_IN_Cnt = bUSB_Length
    USB_BufferReady(HID_BD_IN)          ' Toggle DTS Bit and Turn ownership to SIE
    Return
#endIfMacro-
(*
****************************************************************************
* Name    : HID_RxReport                                                   *
* Require : Value of input argument 'bUSB_Length' should be smaller than   *
*         : the maximum endpoint size responsible for receiving report data*
*         : from USB host for HID CLASS.                                   *
* Input   : pBuffer = Pointer to the starting location of data bytes       *
*         : pLength = Number of bytes To be transferred                    *
* Output  : The number of bytes copied to buffer. Publicly accessible      *
*         : variable bHID_Rpt_Rx_Len is updated with the number of bytes   *
*         : copied to buffer. Once HIDRxReport is called, subsequent       *
*         : retrieval of bHID_Rpt_Rx_Len can be done by calling            *
*         : HID_GetRptRxLength().                                          *
* Purpose : HID_RxReport copies a string of bytes received through         *
*         : USB HID OUT endpoint to a user's specified location. It is a   *
*         : non-blocking function. It does not wait for data if there is   *
*         : no data available. Instead it returns '0' to notify the        *
*         : caller that there is no data available. If the actual number   *
*         : of bytes received is larger than the number of bytes expected  *
*         : (len), only the expected number of bytes specified will be     *
*         : copied to buffer. If the actual number of bytes received is    *
*         : smaller than the number of bytes expected (len), only the      *
*         : actual number of bytes received will be copied to buffer.      *
****************************************************************************
*)
$define HID_RxReport(pBuffer, pLength) '
    wUSB_FSR0 = VarPtr pBuffer         '
    WREG = pLength                     '
    _HIDRxReport

_HIDRxReport Macro- \Byte
    GoSub __HIDRxReport
    Return_Wreg
Endm

#ifMacro- _HIDRxReport, _ReadReport, _ReadArray
__HIDRxReport:
    bUSB_Length = WREG
    bHID_Rpt_Rx_Len = 0
    If tHID_RxIsBusy = False Then
        '
        ' Adjust the expected number of bytes to equal the actual number of bytes received.
        '
        If bUSB_Length > HID_BD_OUT_Cnt Then
            bUSB_Length = HID_BD_OUT_Cnt
        EndIf
        '
        ' Copy data from dual-RAM buffer to user's buffer
        '
        wUSB_FSR1 = VarPtr HID_Report_Out
        While bHID_Rpt_Rx_Len < bUSB_Length
            POSTINC0 = POSTINC1
            Inc bHID_Rpt_Rx_Len
        Wend
        '
        ' Prepare dual-RAM buffer for next OUT transaction
        '
        HID_BD_OUT_Cnt = HID_BUFFER_SIZE
        USB_BufferReady(HID_BD_OUT)                 ' Toggle DTS Bit and Turn ownership to SIE
    EndIf
    WREG = bHID_Rpt_Rx_Len                          ' Return in WREG
    Return
#endIfMacro-
(*
****************************************************************************
* Name    : HID_ReadArray                                                  *
* Purpose : Read an array of bytes - pCount is the number of bytes to read.*
****************************************************************************
*)
    Dim BytesRead As PRODL Shared
    Dim bCount As PRODH Shared

$define HID_ReadArray(pBuffer, pCount) '
    wUSB_FSR0 = VarPtr pBuffer         '
    WREG = pCount                      '
    _ReadArray

_ReadArray Macro-
    GoSub __ReadArray
Endm

#ifMacro- _ReadArray
__ReadArray:
    bCount = WREG
    DisableLowISR()
    While bCount > 0
        If tUSB_Attached = False Then Break
        If tHID_RxIsBusy = False Then                   ' Wait for data
            If bCount > HID_REPORT_BYTES_OUT Then       ' Is bCount > Max report size?
                WREG = HID_REPORT_BYTES_OUT             ' Yes. So
            Else                                        ' Otherwise...
                WREG = bCount
            EndIf
            GoSub __HIDRxReport
            BytesRead = WREG
            bCount = bCount - BytesRead
            wUSB_FSR0 = wUSB_FSR0 + BytesRead
        EndIf
        mUSBService()                   ' Service the USB
    Wend
    EnableLowISR()
    Return
#endIfMacro-
(*
*****************************************************************************
* Name    : HID_WriteArray                                                  *
*                                                                           *
* Input   : pBuffer holds the address of the array to write                 *
*         : pCount is the number of bytes To write                          *
* Purpose : Write an array of bytes -                                       *
* Notes   : If you pass an array which is only dimensioned with 4 bytes     *
*         : (for example, DIM Array[4] as Byte then a 'HID_REPORT_BYTES_IN'   *
*         : sized data packet is still transferred.                         *
*         : This shouldn't be a problem, as the host application should     *
*         : know what to do with the data anyway.                           *
*****************************************************************************
*)
$define HID_WriteArray(pBuffer, pCount) '
    wUSB_FSR0 = VarPtr pBuffer          '
    WREG = pCount                       '
    _WriteArray

_WriteArray Macro-
    GoSub __WriteArray
Endm

#ifMacro- _WriteArray
__WriteArray:
    bCount = WREG
    DisableLowISR()

    While bCount > HID_REPORT_BYTES_IN
        If tUSB_Attached = False Then Break
        USB_WaitForTX()
        WREG = HID_REPORT_BYTES_IN
        GoSub __HIDTxReport
        bCount = bCount - HID_REPORT_BYTES_IN
        wUSB_FSR0 = wUSB_FSR0 + HID_REPORT_BYTES_IN
        mUSBService()             ' Service the USB
    Wend
    If bCount > 0 Then
        USB_WaitForTX()
        WREG = HID_REPORT_BYTES_IN
        GoSub __HIDTxReport
        mUSBService()             ' Service the USB
    EndIf
    EnableLowISR()
    Return
#endIfMacro-
(*
****************************************************************************
* Name    : ReadReport                                                     *
* Purpose : Read a HID report into RXReport buffer                         *
****************************************************************************
*)
$define HID_ReadReport() _ReadReport

_ReadReport Macro- \Byte
    GoSub __ReadReport
    Return_Wreg
Endm

#ifMacro- _ReadReport
__ReadReport:
    DisableLowISR()
    While tHID_RxIsBusy <> False
        If tUSB_Attached = False Then Break
        mUSBService()             ' Service the USB
    Wend
    HID_RxReport(RXReport, HID_REPORT_BYTES_OUT)
    EnableLowISR()
    Return
#endIfMacro-
(*
****************************************************************************
* Name    : WriteReport                                                    *
* Purpose : Write the TXReport buffer                                      *
****************************************************************************
*)
$define HID_WriteReport() _WriteReport

_WriteReport Macro-
    GoSub __WriteReport
Endm

#ifMacro- _WriteReport
__WriteReport:
    DisableLowISR()
    USB_WaitForTX()
    HID_TxReport(TXReport, HID_REPORT_BYTES_IN)
    EnableLowISR()
    Return
#endIfMacro-
'-------------------------------------------------------------------------------------
(*
****************************************************************************
* Name      : putbuf@USB and putbuf@USB@actual                             *
* Purpose   : Replacement for USBout command from RAM                      *
* Input     : GEN holds the EndPoint value (Not Used)                      *
*           : FSR1L/H holds the address of the user buffer                 *
*           : WREG = Length of the user buffer                             *
* Output    : None                                                         *
* Notes     :                                                              *
****************************************************************************
*)
#ifdef Usbout#Ram#Req
Asm-
PutBuf@USB
PutBuf@USB@Actual
Endasm-

#ifndef USBTransmit#Req
    #define USBTransmit#Req
#endif
#ifndef USBSetupForTransmit#Req
    #define USBSetupForTransmit#Req
#endif

    PP0H = WREG                 ' Save the length
    GoSub USBSetupForTransmit
    If tUSB_Attached = False Then __ExitUSB     ' Exit prematurely if the USB becomes detached
'
' Copy data from user's RAM buffer to dual-ram buffer
'
    WREG = PP0H
    Inc WREG
    POSTINC0 = POSTINC1
    Decfsz WREG,f
    Bra ($ - 6)
    Bra _USBTransmit
#endif
(*
*******************************************************************************
* Name      : PutBuf@USB@Auto and PutBuf@USB@Actual@Auto                      *
* Purpose   : Replacement for USBout command from RAM with Auto length detect *
* Input     : GEN holds the EndPoint value (Not Used)                         *
*           : FSR1L/H holds the address of the user buffer                    *
* Output    : None                                                            *
* Notes     :                                                                 *
*******************************************************************************
*)
#ifdef Usbout#Ram#Auto#Req
Asm-
PutBuf@USB@Auto
PutBuf@USB@Actual@Auto
Endasm-
#ifndef USBTransmit#Req
    #define USBTransmit#Req
#endif
#ifndef USBSetupForTransmit#Req
    #define USBSetupForTransmit#Req
#endif

    GoSub USBSetupForTransmit
    If tUSB_Attached = False Then __ExitUSB     ' Exit prematurely if the USB becomes detached
'
' Copy data from user's RAM buffer to dual-ram buffer
'
    PP0H = 0
    WREG = POSTINC1
    Bz ($ + 8)
    POSTINC0 = WREG
    Incfsz PP0H,f
    Bra ($ - 8)
    Bra _USBTransmit
#endif
(*
***********************************************************************************
* Name      : PutMem@USB                                                          *
* Purpose   : Replacement for USBout command from code memory                     *
* Input     : GEN holds the EndPoint value (Not Used)                             *
*           : TBLPTRL/H holds the (16-bit) address of the user code memory buffer *
*           : WREG = Length of the user buffer                                    *
* Output    : None                                                                *
* Notes     :                                                                     *
***********************************************************************************
*)
#ifdef Usbout#Mem#Req
Asm
PutMem@USB
EndAsm
#ifndef USBTransmit#Req
#define USBTransmit#Req
#endif
#ifndef USBSetupForTransmit#Req
#define USBSetupForTransmit#Req
#endif
    PP0H = WREG                 ' Save the length
    GoSub USBSetupForTransmit
    If tUSB_Attached = False Then __ExitUSB     ' Exit prematurely if the USB becomes detached
'
' Copy data from user's (code memory) buffer to dual-ram buffer
'
    WREG = PP0H
    Inc WREG
    Tblrd*+                             ' Read from code memory
    POSTINC0 = TABLAT
    Decfsz WREG,f
    Bra ($ - 8)
    Bra _USBTransmit
#endif
(*
***************************************************************************************
* Name      : PutMem@USB@Auto                                                         *
* Purpose   : Replacement for USBout command from code memory with auto length detect *
* Input     : GEN holds the EndPoint value (Not Used)                                 *
*           : TBLPTRL/H holds the (16-bit) address of the user code memory buffer     *
* Output    : None                                                                    *
* Notes     :                                                                         *
***************************************************************************************
*)
#ifdef Usbout#Mem#Auto#Req
Asm
PutMem@USB@Auto
EndAsm
#ifndef USBTransmit#Req
#define USBTransmit#Req
#endif
#ifndef USBSetupForTransmit#Req
#define USBSetupForTransmit#Req
#endif
    GoSub USBSetupForTransmit
    If tUSB_Attached = False Then __ExitUSB  ' Exit prematurely if the USB becomes detached
'
' Copy data from user's (code memory) buffer to dual-ram buffer
'
    PP0H = 0
    Tblrd*+                             ' Read from code memory
    WREG = TABLAT
    Bz ($ + 8)
    POSTINC0 = WREG
    Incfsz PP0H,f
    Bra ($ - 10)

    Bra _USBTransmit
#endif
(*
****************************************************************************
* Name      : USBSetupForTransmit                                          *
* Purpose   : Setup before transmitting buffer                             *
****************************************************************************
*)
#ifdef USBSetupForTransmit#Req
USBSetupForTransmit:
    DisableLowISR()
    USB_WaitForTX()                     ' Wait for the Dual Port RAM to become available
    Clear HID_Report_In                 ' Clear the USB dual port buffer

    wUSB_FSR0 = VarPtr HID_Report_In    ' Point to the dual port RAM before returning
    Return
#endif
(*
****************************************************************************
* Name      : _USBTransmit                                                 *
* Purpose   : Transmit the contents of the USB buffer                      *
****************************************************************************
*)
#ifdef USBTransmit#Req
_USBTransmit:
    HID_BD_IN_Cnt = HID_REPORT_BYTES_IN
    USB_BufferReady(HID_BD_IN)              ' Toggle DTS Bit and Turn ownership to SIE
    mUSBService()                           ' Service the USB
__ExitUSB:
    EnableLowISR()
    Return
#endif

'-------------------------------------------------------------------------------------
(*
*******************************************************************************
* Name      : Get@USB@Auto and GetBuf@USB@Auto                                *
* Purpose   : Replacement for USBin command from RAM with Auto length detect  *
* Input     : GEN holds the EndPoint value (Not Used)                         *
*           : FSR1L/H holds the address of the user buffer                    *
* Output    : None                                                            *
* Notes     :                                                                 *
*******************************************************************************
*)
#ifdef Usbin#Buffer#Auto#Req
Asm-
Get@USB@Auto
GetBuf@USB@Auto
Endasm-
    DisableLowISR()
    While tHID_RxIsBusy <> False
        If tUSB_Attached = False Then __ExitUSBReceive
        mUSBService()                     ' Service the USB
    Wend
    'wUSB_FSR0 = VarPtr RXReport
    'HID_RxReport(HID_REPORT_BYTES_OUT)
    If tHID_RxIsBusy = False Then
        ' Adjust the expected number of bytes to equal the actual number of bytes received.
        'If bUSB_Length > HID_BD_OUT_Cnt Then
        '    bUSB_Length = HID_BD_OUT_Cnt
        'EndIf
        '
        ' Copy data from dual-ram buffer to user's buffer
        '
        wUSB_FSR0 = VarPtr HID_Report_Out                ' Point to the dual port USB buffer
        'While hid_rpt_rx_len < bUSB_Length
        While 1 = 1
            PP0H = POSTINC0                             ' Extract a byte from the USB buffer
            If PP0H = 0 Then Break                      ' Exit when a Null found
            POSTINC1 = PP0H                             ' Copy the byte to the user buffer
            'Inc hid_rpt_rx_len
        Wend
        '
        ' Prepare dual-ram buffer for next out transaction
        '
        HID_BD_OUT_Cnt = HID_BUFFER_SIZE
        USB_BufferReady(HID_BD_OUT)                     ' Toggle DTS Bit and Turn ownership to SIE
    EndIf
__ExitUSBReceive:
    EnableLowISR()
    Return
#endif
(*
*******************************************************************************
* Name      : Get@USB@Auto and GetBuf@USB@Auto                                *
* Purpose   : Replacement for USBin command from RAM with Auto length detect  *
* Input     : GEN holds the EndPoint value (Not Used)                         *
*           : FSR1L/H holds the address of the user buffer                    *
*           : WREG holds the amount of bytes to receive                       *
* Output    : None                                                            *
* Notes     :                                                                 *
*******************************************************************************
*)
#ifdef Usbinbuf#Req
    #ifndef Usbin#Req
        #define Usbin#Req
    #endif
#endif
#ifdef Usbin#Req
Asm-
Get@USB
GetBuf@USB
Endasm-
    PP0H = WREG                         ' Save Length
    DisableLowISR()
    While tHID_RxIsBusy <> False
        If tUSB_Attached = False Then __ExitUSBReceive_
        mUSBService()                   ' Service the USB
    Wend
    If tHID_RxIsBusy = False Then
        ' Adjust the expected number of bytes to equal the actual number of bytes received.
        If PP0H > HID_BD_OUT_Cnt Then
            PP0H = HID_BD_OUT_Cnt
        EndIf
        '
        ' Copy data from dual-ram buffer to user's buffer
        '
        wUSB_FSR0 = VarPtr HID_Report_Out               ' Point to the dual port USB buffer
        While WREG < PP0H
            POSTINC1 = POSTINC0                         ' Copy the byte to the user buffer
            Inc PP0H
        Wend
        '
        ' Prepare dual-ram buffer for next OUT transaction
        '
        HID_BD_OUT_Cnt = HID_BUFFER_SIZE
        USB_BufferReady(HID_BD_OUT)                     ' Toggle DTS Bit and Turn ownership to SIE
    EndIf
__ExitUSBReceive_:
    EnableLowISR()
    Return
#endif
OverHIDSubs:

